/*  GSMP - General Sound Manipulation Program
 *  Copyright (C) 2000-2002 Ren Rebe and Valentin Ziegler
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* Short Description:
 *   Normal graphic view. Currently moving towards the over-lays ...
 */

#ifndef GSMP_NORMAL_VIEW
#define GSMP_NORMAL_VIEW

// GSML stuff:

#include "include.hh"
#include "typedefs.hh"
#include "Basics.hh"

#include "WavePlot.hh"

// stl containers
#include <vector>
#include <deque>

// GSMP
#include "ToolkitBasics.hh"
#include "TrackInfo.hh" 
#include "Canvas.hh"

// gtk--
#include <gtk--/drawingarea.h> 
#include <gtk--/box.h>
#include <gtk--/ruler.h> 
#include <gtk--/scrollbar.h>
#include <gtk--/adjustment.h>

#define SCROLL_CACHE_FACTOR 10
#define SMOTH_SCROLL_FACTOR 4
#define AUTO_SCOLL_ACCEL 6

namespace GSMP
{
  using namespace GSML;
  
  /* Documentation / Brainstorming ...
   * Must be able to:
   *  - 1. render waveforms, selection, markers, parameter-lines in a nice way
   *  - 2. have good speed
   *  - 3. (other overlays, too)
   */
  
  /* This is a canvas style replacement for the gnome-canvas. We do
   * not use them), because the C++ wraper is not powerfull enough and
   * we do not want to link against all the GNOME stuff.
   *
   * This now becomes the basics of an whole graphic sub-system -
   * since all current free implementations are not that
   * satisfying. And after reading some theoritcal basics I want to
   * try a really object oriented approach using also somem generic
   * helpers that might help in other cases, too.
   * 
   * First try Canvas, using long integers (unlike gnome using floats)
   * and without automatic proportinal zooming (like the gnome one
   * does). This is not needed (and counter-productive here ?!).
   *
   * The canvas's zoom is only passed to the items - which must do the
   * zooming themselfes (incl. possible resizements). - The difference
   * to the GNOME zoom is that not all gets proportinally
   * zoomed. E.g. lines don't get thicker - and no world-coordinate is
   * changed. This is what we need. For us the zoom is simply a
   * message for the items to maybe do s.th. The waveforms will change
   * it's length, but not its height, the markers text will be of the
   * same height, but the markers must appear zoomed to be at the same
   * place over the wavefrom.
   *
   * The items's layer indicates in which order the items are drawn,
   * and the transparent flag is a help for the canvas to do
   * optimizations which items must be rendered and which can be
   * skipped (eg. because a non-transparent is over it ...).
   *
   * The canvas is a GUI / OS independent canvas which renders the
   * content into a generic pixmap - so plugins can provide some
   * cavas_item without having to care on which toolkit / platforn it
   * is running.
   */
  
  class WaveFormItem : public GsmpCanvasItem
  {
  public:
    WaveFormItem (GsmpCanvas* i_canvas, GSML::WavePlotManager* i_manager, track_index i_track);
    virtual ~WaveFormItem ();
    
    // GsmpCanvasItem methods
    void SetZoom (int n_zoom);
    
    virtual void Render (GsmpPixmap* pixmap, const canvas_area& item_area,
			 const canvas_point& dest);
    
    // none of the user input methods needed ...
    
  protected:
    
    void UpdateArea ();
    void ConfigWavePlot ();
    
    void SlotProgressUpdated (sample_index start, sample_index length);
    void SlotWavePlotChanged (MinMaxAvgWavePlot* n_plot);
    
    int m_zoom;
    
    GSML::WavePlotManager* m_manager;
    GSML::track_index m_track;
    
    GSML::WavePlotConfig m_wp_config;
    GSML::WavePlotTracer<GSML::MinMaxAvgWavePlot>* m_waveplot;
  };
  
  // ***
  
  class SelectionItem : public GsmpCanvasItem
  {
  public:
    SelectionItem (GsmpCanvas* i_canvas, GSML::WavePlotManager* i_manager, track_index i_track);
    virtual ~SelectionItem ();
    
    // GsmpCanvasItem methods
    void SetZoom (int n_zoom);
    virtual void Render (GsmpPixmap* pixmap, const canvas_area& item_area,
			 const canvas_point& dest);
    
    // user input methods we implement
    virtual bool PointerMotion (const canvas_point& p);
    virtual bool ButtonPress (int button);
    virtual bool BottonRelease (int button);
    
  protected:
    
    GSML::WavePlotManager* m_manager;
    GSML::track_index m_track;
    
    GSML::WavePlotConfig m_wp_config;
    GSML::WavePlotTracer<GSML::MinMaxAvgWavePlot>* m_waveplot;
  };

  // ***
  
  class NormalView : public Gtk::Table, public ViewImpl
  {
  public:
    NormalView (View* i_view);
    NormalView (View* i_view, const NormalView& src);
    NormalView (const NormalView& src);
    virtual ~NormalView ();
  
    virtual void SetStart (sample_index n_start);
    virtual sample_index Start () const;
    
    virtual void SetZoom (sample_index n_zoom);
    virtual void SetZoomAll ();
    virtual int Zoom () const;
  
    virtual sample_index Width () const;
    virtual sample_index Length () const;
  
    virtual void SetTrackVisible (track_index track_no, bool status);
    virtual bool IsTrackVisible (track_index track_no) const;
    
    virtual sample_index Pointer () const;
    
    virtual ViewImpl* Clone (View* n_view);
    
    void Debug ();
    
  protected:
    
    /********************
     * internal helpers *
     ********************/
    void initialize ();

    void clear_table ();
    void rebuild_table ();
    
    void set_start (sample_index n_start);
    void set_zoom (int n_zoom);
    
    // Data
    void SlotCanvasAreaResized (const canvas_area& n_area);
    void SlotCanvasViewportModified (const canvas_area& n_area);
    
    void SlotLayerChanged (DataLayer* n_layer);
    
    void SlotAdjustmentValueChanged ();
    
    void SlotGuiRefresh (); // clock signal for delayed GUI refresh 
    
    void SlotButtonPress (int button);
    
    bool m_auto_zoom_all;        // auto zoom all after layer change ?
    bool m_viewport_internal_mod;
    
    struct view_stuff {
      TrackInfo* track_info;
      WaveFormItem* waveform;
    };
    
    GsmpCanvas m_canvas;
    deque<view_stuff> m_views;
  
    WavePlotManager* m_manager;
    Signal0<void>* m_timeout;
  
    Gtk::HScrollbar m_scrollbar;
    Gtk::Adjustment m_adjustment;
  
    GSMP::Menu::MenuCollection* m_menu;
    MenuGtk* menu_popup;
  };

} // namespace GSMP
#endif // GSMP_NORMAL_VIEW

#ifdef OLD_CODE
// sub class for a single track
class NormalTrackView : public Gtk::DrawingArea
{
public:
  NormalTrackView (WavePlotTracer<MinMaxAvgWavePlot>* i_plot, const WavePlotConfig* i_config,
		   NormalView* i_parent, int i_width, bool i_master = false); // master will draw additional graphic stuff (marker heads ...)
  ~NormalTrackView ();
  
  void SetStartDelta (int n_delta, bool force_redaw = false);
  
  void SetTempSelection (sample_index n_start, sample_index n_end);
  void RemoveTempSelection ();
  
  void SetRefresh (); // only temp. debug hack (context menu call ...)
  
  int Width () const;
  int Height () const;
  
  void SlotGUIRefresh ();
    
  void Debug () const;
    
private:
  
  void add_refresh (int n_start, int n_end);
  void add_refresh_from_real (sample_index start, sample_index length);
  
  void add_refresh_all ();
  void refresh_clear ();
  
  virtual void draw_wave (); // internal to draw the wave
    
  const MinMaxAvgWavePlot::Progress* waveform_progress;
  int waveform_middle;
  int waveform_vscale;
  int waveform_dBfs;
  
  virtual void prepare_waveform (const MinMaxAvgWavePlot::Progress* n_progress);
  virtual void draw_waveform (int view_pos, Gdk_GC& gc, int plot_pos);
  virtual void draw_grid (int view_pos, Gdk_GC& gc, bool draw_vert);
  virtual void draw_marker (int view_pos, Gdk_GC& gc);
  
  void connect_selection ();
  void connect_markers ();
  
  // Signal Slots:
  void SlotWavePlotChanged (MinMaxAvgWavePlot* n_plot);
  void SlotWavePlotUpdated (sample_index start, sample_index length);
  
  SigC::Connection connection_selection_selected;
  void SlotSelectionSelected (sample_index, sample_index);
  SigC::Connection connection_selection_unselected;
  void SlotSelectionUnselected (sample_index, sample_index);
  SigC::Connection connection_selection_active_changed;
  void SlotSelectionActiveChanged (bool n_state);
  
  SigC::Connection connection_marker_added;
  void SlotMarkerAdded (const string& n_name, sample_index n_pos);
  SigC::Connection connection_marker_deleted;
  void SlotMarkerDeleted (const string& n_name, sample_index n_pos);
  SigC::Connection connection_marker_changed;
  void SlotMarkerChanged (const string& n_name, sample_index o_pos, sample_index n_pos);
  
  // gtk-- event-impl overwrites ...
  virtual void realize_impl ();
  virtual gint configure_event_impl (GdkEventConfigure* event);
  virtual gint expose_event_impl (GdkEventExpose* event);
  virtual gint button_press_event_impl (GdkEventButton* event);
  virtual gint button_release_event_impl (GdkEventButton* event);
  virtual gint motion_notify_event_impl (GdkEventMotion* event);
  
  // Gtk-- stuff
  Gdk_Window win;
  
  Gdk_Pixmap pixmap; // the cached image data
  
  Gdk_GC back_gc;  // the normal background
  Gdk_GC back_sel_gc;  //  background for selections
  Gdk_GC back_temp_sel_gc;  //  background for temp selections
  Gdk_GC back_undef_gc; // undef (not yet loaded) background
  Gdk_GC wave_gc;  // for the wave
  Gdk_GC wave_sel_gc;  // for the wave
  Gdk_GC wave_temp_sel_gc;  // for the wave
  Gdk_GC grid_gc;  // for the grid
  Gdk_GC marker_gc;  // for the markers
  
  // internal states
  int m_width; // current width and height of our window
  int m_height;
  
  int m_start_delta; // the start in plot we should display
  int m_internal_delta; // our start in plot of cache pixmap
  
  int m_cache_length; // the length of our cache pixmap
  
  sample_index m_temp_sel_start;
  sample_index m_temp_sel_end;
  bool m_use_temp_sel;
  
  SigC::Threads::Mutex refresh_lock;
  bool refresh_needed;
  vector <bool> refresh_region;
  
  // GSML Stuff
  WavePlotTracer<MinMaxAvgWavePlot>* m_waveplot;
  const WavePlotConfig* m_config;
  NormalView* m_parent;
  bool m_master;
  
  MinMaxAvgWavePlot* m_current_plot;
  TrackLayer::Selection* m_current_sel;
  DataLayer::Markers* m_current_markers;
};
#endif
